package org.livee.data.query.resolve;
/*
 * Copyright (c) 2007 Hideya Ochiai,  All rights reserved.
 * 
 * Permission of redistribution and use in source and binary forms, 
 * with or without modification, are granted, free of charge, to any person 
 * obtaining the copy of this software under the following conditions:
 * 
 *  1. Any copies of this source code must include the above copyright notice,
 *  this permission notice and the following statement without modification 
 *  except possible additions of other copyright notices. 
 * 
 *  2. Redistributions of the binary code must involve the copy of the above 
 *  copyright notice, this permission notice and the following statement 
 *  in documents and/or materials provided with the distribution.  
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

import org.livee.data.exception.LiveEException;
import org.livee.data.exception.NoServerResolvedLiveEException;
import org.livee.data.expr.AbstractData;
import org.livee.data.expr.AbstractParser;
import org.livee.data.expr.ParserFactory;
import org.livee.data.soap.client.Admin200703Service;
import org.livee.data.soap.client.Admin200703ServiceLocator;
import org.livee.data.soap.client.Admin200703_PortType;
import org.livee.data.soap.client.DataProvider200703Service;
import org.livee.data.soap.client.DataProvider200703ServiceLocator;
import org.livee.data.soap.client.DataProvider200703_PortType;
import org.livee.util.Xml;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * @author Hideya Ochiai, The University of Tokyo
 * Created: before 2008-01-01
 * Last Updated: 2008-11-05 
 */
public class RetrieveManager {

	public class RetrieveManagerCache extends Thread {
		
		public class CacheEntry {
			public AbstractData queryResponse;
			public long cachedTime;
			public CacheEntry(AbstractData qr,long time){
				queryResponse=qr;
				cachedTime=time;
			}
		}
		
		java.util.Map m_CacheMap;
		int m_CacheValidityTime;
		int m_MaxCacheSize;
		java.util.regex.Pattern m_CachePredicate;
		
		public RetrieveManagerCache(){
			m_CacheMap=new java.util.Hashtable();
			
			org.livee.data.tools.ConfigManager config=org.livee.data.tools.ConfigManager.getInstance();
			int validityTime=config.getIntParameter("/c:liveeConfiguration/c:search/c:retrieveManager/c:cacheValidityTime/text()");
			int cacheSize=config.getIntParameter("/c:liveeConfiguration/c:search/c:retrieveManager/c:cacheSize/text()");
			String cachePredicate=config.getStringParameter("/c:liveeConfiguration/c:search/c:retrieveManager/c:cachePredicate/text()");
			if(validityTime>0){
				m_CacheValidityTime=validityTime*1000;				
			}else{
				m_CacheValidityTime=60000; // recommended default 60sec
			}
			if(cacheSize>0){
				m_MaxCacheSize=cacheSize;
			}else{
				m_MaxCacheSize=10; // recommended default 10
			}
			
			try{
				m_CachePredicate=java.util.regex.Pattern.compile(cachePredicate);
			}
			catch(Exception e){
				m_CachePredicate=java.util.regex.Pattern.compile(".*");
				// TODO: predicate warning
			}
			
			start();
		}
		
		public AbstractData getCachedData(String key){
			Object o=m_CacheMap.get(key);
			if(o!=null){
				CacheEntry entry=(CacheEntry)o;
				//System.out.println("MR GET SUCCEEDED: "+key);
				return entry.queryResponse;
			}
			return null;
		}
		
		public void setData(String key, AbstractData sensorCollection, long current){
			CacheEntry entry=new CacheEntry(sensorCollection,current);
			if(m_CacheMap.size()<m_MaxCacheSize){
				//System.out.println("MR PUT SUCCEEDED: "+key);
				synchronized(m_CacheMap){
					m_CacheMap.put(key, entry);
				}
			}
		}
		
		public void run(){
			// Cache removal 
			while(true){
				java.util.ArrayList keyset=new java.util.ArrayList();
				synchronized(m_CacheMap){
					keyset.addAll(m_CacheMap.keySet());
				}
				for(int i=0;i<keyset.size();i++){
					long current=System.currentTimeMillis();
					String key=(String)keyset.get(i);
					synchronized(m_CacheMap){
						CacheEntry entry=(CacheEntry)m_CacheMap.get(key);
						if((current-m_CacheValidityTime)>entry.cachedTime || current<entry.cachedTime){
							m_CacheMap.remove(key);
						}
					}
					try{
						Thread.sleep(100);
					}
					catch(Exception e){}
				}
			
				try{
					Thread.sleep(5000);
				}catch(Exception e){
					
				}	
			}
		}
	}
	
	
	static RetrieveManager m_Singleton=null;
	RetrieveManagerCache m_Cache=null;
	
	RetrieveManager(){
		m_Cache=new RetrieveManagerCache();
		m_Singleton=this;
	}
	
	public static RetrieveManager getInstance(){
		if(m_Singleton==null){
			new RetrieveManager();
		}
		return m_Singleton;
	}
	
	public AbstractData retrieve(IRetriever retriever) throws LiveEException {

		// Data Cache Check
		String key=retriever.generateKey();
		AbstractData sensorCollection=m_Cache.getCachedData(key);
		if(sensorCollection!=null){
			return sensorCollection;
		}

		// Server Resolution
		String query_str=retriever.getQueryString();
		AbstractData query=retriever.getQuery();
		QueryResolver resolver=QueryResolver.getInstance();
		AbstractData queryResponse=resolver.resolve(query,query_str);
		if(queryResponse==null){
			throw new NoServerResolvedLiveEException(query);
		} 
		
		// Data Retrieval & Merge
		boolean bTried=false;
		java.util.Iterator itr=queryResponse.getChildren().iterator();
		while(itr.hasNext()){
			AbstractData adminSite=(AbstractData)itr.next();
			AbstractData siteSensorCollection=null;
			if(adminSite!=null && adminSite.getDataType().equals("AdminSite")){
				bTried=true;
				siteSensorCollection=retriever.retrieveSite(adminSite);
				
				if(siteSensorCollection!=null){
					if(sensorCollection==null){
						sensorCollection=siteSensorCollection;
					}else{
						java.util.Iterator sensorItr=siteSensorCollection.getChildren().iterator();
						while(sensorItr.hasNext()){
							sensorCollection.append((AbstractData)sensorItr.next());
						}
					}
				}
			}
		}
		if(!bTried){
			throw new NoServerResolvedLiveEException(query);			
		}
		
		// Cache sensorCollection
		m_Cache.setData(key, sensorCollection, System.currentTimeMillis());
		return sensorCollection;
	}
}
